use std::env;
use std::error::Error;
use std::net::{TcpListener, TcpStream};
use std::sync::Arc;

use rcgen::generate_simple_self_signed;
use rustls::crypto::aws_lc_rs::sign::any_supported_type;
use rustls::pki_types::pem::PemObject;
use rustls::pki_types::{CertificateDer, PrivateKeyDer};
use rustls::server::ResolvesServerCertUsingSni;
use rustls::sign::CertifiedKey;
use rustls::{ServerConfig, ServerConnection, StreamOwned};
use threadpool::ThreadPool;

fn accept_client(stream: TcpStream, config: Arc<ServerConfig>, sni_domain: String) -> Result<(), Box<dyn Error>> {
    let connection = ServerConnection::new(config)?;

    let mut stream = StreamOwned::new(connection, stream);

    while stream.conn.is_handshaking() {
        stream.conn.complete_io(&mut stream.sock)?;
    }

    if stream.conn.server_name() != Some(&sni_domain) {
        return Err("unexpected sni domain".into());
    }

    

    Ok(())
}

fn main() -> Result<(), Box<dyn Error>> {
    let mut args = env::args();

    args.next();

    let sni_domain = args
        .next()
        .expect("missing sni domain argument");
    let host = args
        .next()
        .expect("missing host argument");

    let certified_key = generate_simple_self_signed(vec![sni_domain.to_string()]).unwrap();
    let certified_key = CertifiedKey::new(
        vec![certified_key.cert.into()], 
        any_supported_type(
            &PrivateKeyDer::from_pem_slice(certified_key.key_pair.serialize_pem().as_bytes())?
        )?
    );

    let mut resolver = ResolvesServerCertUsingSni::new();
    resolver.add(&sni_domain, certified_key)?;

    let config = Arc::new(ServerConfig::builder()
        .with_no_client_auth()
        .with_cert_resolver(Arc::new(resolver)));

    let threadpool = ThreadPool::new(10);

    let listener = TcpListener::bind(host).unwrap();

    for stream in listener.incoming() {
        let stream = stream.expect("listener got broken");

        threadpool.execute({
            let config = config.clone();
            let sni_domain = sni_domain.clone();

            move || {
                if let Err(e) = accept_client(stream, config, sni_domain) {
                    println!("error connection: {e:?}");
                }
            }
        });
    }

    Ok(())
}